@@ -138,25 +138,6 @@ _p11_library_get_thread_local (void)
 }
 #endif
 
-#if defined(HAVE_DECL___REGISTER_ATFORK) && !HAVE_DECL___REGISTER_ATFORK
-extern int __register_atfork (void (*prepare) (void), void (*parent) (void),
-			      void (*child) (void), void *dso_handle);
-#endif /* HAVE_DECL___REGISTER_ATFORK */
-
-#ifdef HAVE___REGISTER_ATFORK
-
-extern void *__dso_handle;
-
-#define p11_register_atfork(a,b,c) \
-	(__register_atfork((a),(b),(c),__dso_handle))
-
-#else
-
-#define p11_register_atfork(a,b,c) \
-	(pthread_atfork((a),(b),(c)))
-
-#endif /* HAVE___REGISTER_ATFORK */
-
 static void
 count_forks (void)
 {
@@ -179,7 +160,7 @@ p11_library_init_impl (void)
 	p11_message_locale = newlocale (LC_ALL_MASK, "POSIX", (locale_t) 0);
 #endif
 
-	p11_register_atfork (NULL, NULL, count_forks);
+	pthread_atfork (NULL, NULL, count_forks);
 }
 
 void
