@@ -44,6 +44,7 @@
 #include "private.h"
 #include "rpc.h"
 #include "rpc-message.h"
+#include "path.h"
 
 #include <sys/types.h>
 
@@ -51,6 +52,7 @@
 #include <errno.h>
 #include <fcntl.h>
 #include <stdint.h>
+#include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
 
@@ -64,6 +66,8 @@
 #endif
 
 #ifdef OS_WIN32
+#include <process.h>
+#include <signal.h>
 #include <winsock2.h>
 #ifndef EWOULDBLOCK
 #define EWOULDBLOCK WSAEWOULDBLOCK
@@ -75,8 +79,11 @@
 #endif
 
 typedef struct {
-	/* Never changes */
-	int fd;
+	/* Never changes.  On Unix, these are identical, as it is
+	 * backed by a socket.  On Windows, it is another file
+	 * descriptor, as they are backed by two pipes */
+	int read_fd;
+	int write_fd;
 
 	/* Protected by the lock */
 	p11_mutex_t write_lock;
@@ -86,6 +93,10 @@ typedef struct {
 
 	/* This data is protected by read mutex */
 	p11_mutex_t read_lock;
+#ifdef OS_UNIX
+        /* Signalled when read_code changes */
+        p11_cond_t read_code_cond;
+#endif
 	bool read_creds;
 	uint32_t read_code;
 	uint32_t read_olen;
@@ -100,7 +111,8 @@ rpc_socket_new (int fd)
 	sock = calloc (1, sizeof (rpc_socket));
 	return_val_if_fail (sock != NULL, NULL);
 
-	sock->fd = fd;
+	sock->read_fd = fd;
+	sock->write_fd = fd;
 	sock->last_code = 0x10;
 	sock->read_creds = false;
 	sock->sent_creds = false;
@@ -109,6 +121,10 @@ rpc_socket_new (int fd)
 	p11_mutex_init (&sock->write_lock);
 	p11_mutex_init (&sock->read_lock);
 
+#ifdef OS_UNIX
+        p11_cond_init (&sock->read_code_cond);
+#endif
+
 	return sock;
 }
 
@@ -129,7 +145,7 @@ static bool
 rpc_socket_is_open (rpc_socket *sock)
 {
 	assert (sock != NULL);
-	return sock->fd >= 0;
+	return sock->read_fd >= 0;
 }
 #endif
 
@@ -137,9 +153,14 @@ static void
 rpc_socket_close (rpc_socket *sock)
 {
 	assert (sock != NULL);
-	if (sock->fd != -1)
-		close (sock->fd);
-	sock->fd = -1;
+	if (sock->read_fd != -1)
+		close (sock->read_fd);
+	sock->read_fd = -1;
+#ifdef OS_WIN32
+	if (sock->write_fd != -1)
+		close (sock->write_fd);
+	sock->write_fd = -1;
+#endif
 }
 
 static void
@@ -163,6 +184,9 @@ rpc_socket_unref (rpc_socket *sock)
 	rpc_socket_close (sock);
 	p11_mutex_uninit (&sock->write_lock);
 	p11_mutex_uninit (&sock->read_lock);
+#ifdef OS_UNIX
+        p11_cond_uninit (&sock->read_code_cond);
+#endif
 	free (sock);
 }
 
@@ -234,7 +258,7 @@ rpc_socket_write_inlock (rpc_socket *sock,
 
 	/* Place holder byte, will later carry unix credentials (on some systems) */
 	if (!sock->sent_creds) {
-		if (write_all (sock->fd, &dummy, 1) != 1) {
+		if (write_all (sock->write_fd, &dummy, 1) != 1) {
 			p11_message_err (errno, "couldn't send socket credentials");
 			return CKR_DEVICE_ERROR;
 		}
@@ -245,9 +269,9 @@ rpc_socket_write_inlock (rpc_socket *sock,
 	p11_rpc_buffer_encode_uint32 (header + 4, options->len);
 	p11_rpc_buffer_encode_uint32 (header + 8, buffer->len);
 
-	if (!write_all (sock->fd, header, 12) ||
-	    !write_all (sock->fd, options->data, options->len) ||
-	    !write_all (sock->fd, buffer->data, buffer->len))
+	if (!write_all (sock->write_fd, header, 12) ||
+	    !write_all (sock->write_fd, options->data, options->len) ||
+	    !write_all (sock->write_fd, buffer->data, buffer->len))
 		return CKR_DEVICE_ERROR;
 
 	return CKR_OK;
@@ -344,6 +368,24 @@ p11_rpc_transport_write (int fd,
 	return status;
 }
 
+static void
+rpc_socket_set_read_code_inlock (rpc_socket *sock,
+                                 int code)
+{
+        sock->read_code = code;
+#ifdef OS_UNIX
+        p11_cond_broadcast (&sock->read_code_cond);
+#endif
+}
+
+#ifdef OS_UNIX
+static void
+rpc_socket_wait_for_read_code_change_inlock (rpc_socket *sock)
+{
+        p11_cond_wait (&sock->read_code_cond, &sock->read_lock);
+}
+#endif
+
 static int
 rpc_socket_read (rpc_socket *sock,
                  int *code,
@@ -352,7 +394,10 @@ rpc_socket_read (rpc_socket *sock,
 	CK_RV ret = CKR_DEVICE_ERROR;
 	unsigned char header[12];
 	unsigned char dummy;
-	fd_set rfds;
+#ifdef OS_WIN32
+	HANDLE handle;
+	DWORD mode;
+#endif
 
 	assert (code != NULL);
 	assert (buffer != NULL);
@@ -365,7 +410,7 @@ rpc_socket_read (rpc_socket *sock,
 	p11_mutex_lock (&sock->read_lock);
 
 	if (!sock->read_creds) {
-		if (read_all (sock->fd, &dummy, 1) != 1) {
+		if (read_all (sock->read_fd, &dummy, 1) != 1) {
 			p11_mutex_unlock (&sock->read_lock);
 			return CKR_DEVICE_ERROR;
 		}
@@ -375,11 +420,11 @@ rpc_socket_read (rpc_socket *sock,
 	for (;;) {
 		/* No message header has been read yet? ... read one in */
 		if (sock->read_code == 0) {
-			if (!read_all (sock->fd, header, 12))
+			if (!read_all (sock->read_fd, header, 12))
 				break;
 
 			/* Decode and check the message header */
-			sock->read_code = p11_rpc_buffer_decode_uint32 (header);
+			rpc_socket_set_read_code_inlock (sock, p11_rpc_buffer_decode_uint32 (header));
 			sock->read_olen = p11_rpc_buffer_decode_uint32 (header + 4);
 			sock->read_dlen = p11_rpc_buffer_decode_uint32 (header + 8);
 			if (sock->read_code == 0) {
@@ -399,15 +444,15 @@ rpc_socket_read (rpc_socket *sock,
 			}
 
 			/* Read in the the options first, and then data */
-			if (!read_all (sock->fd, buffer->data, sock->read_olen) ||
-			    !read_all (sock->fd, buffer->data, sock->read_dlen))
+			if (!read_all (sock->read_fd, buffer->data, sock->read_olen) ||
+			    !read_all (sock->read_fd, buffer->data, sock->read_dlen))
 				break;
 
 			buffer->len = sock->read_dlen;
 			*code = sock->read_code;
 
 			/* Yay, we got our data, off we go */
-			sock->read_code = 0;
+                        rpc_socket_set_read_code_inlock (sock, 0);
 			sock->read_olen = 0;
 			sock->read_dlen = 0;
 			ret = CKR_OK;
@@ -417,15 +462,18 @@ rpc_socket_read (rpc_socket *sock,
 		/* Give another thread the chance to read data for this header */
 		if (sock->read_code != 0) {
 			p11_debug ("received header in wrong thread");
-			p11_mutex_unlock (&sock->read_lock);
 
+#ifdef OS_UNIX
+                        rpc_socket_wait_for_read_code_change_inlock (sock);
+#endif
+#ifdef OS_WIN32
 			/* Used as a simple wait */
-			FD_ZERO (&rfds);
-			FD_SET (sock->fd, &rfds);
-			if (select (sock->fd + 1, &rfds, NULL, NULL, NULL) < 0)
-				p11_message ("couldn't use select to wait on rpc socket");
-
+			p11_mutex_unlock (&sock->read_lock);
+			handle = (HANDLE) _get_osfhandle (sock->read_fd);
+			if (!ReadFile (handle, NULL, 0, &mode, NULL))
+				p11_message ("couldn't use select to wait on rpc pipe");
 			p11_mutex_lock (&sock->read_lock);
+#endif
 		}
 	}
 
@@ -610,8 +658,12 @@ rpc_transport_buffer (p11_rpc_client_vtable *vtable,
 	/* Get the next socket reply code */
 	call_code = sock->last_code++;
 
-	if (sock->fd == -1)
+	if (sock->read_fd == -1)
 		rv = CKR_DEVICE_ERROR;
+#ifdef OS_WIN32
+	if (sock->write_fd == -1)
+		rv = CKR_DEVICE_ERROR;
+#endif
 	if (rv == CKR_OK)
 		rv = rpc_socket_write_inlock (sock, call_code, &rpc->options, request);
 
@@ -624,11 +676,18 @@ rpc_transport_buffer (p11_rpc_client_vtable *vtable,
 		p11_mutex_lock (&sock->write_lock);
 	}
 
-	if (rv != CKR_OK && sock->fd != -1) {
+	if (rv != CKR_OK && sock->read_fd != -1) {
 		p11_message ("closing socket due to protocol failure");
-		close (sock->fd);
-		sock->fd = -1;
+		close (sock->read_fd);
+		sock->read_fd = -1;
 	}
+#ifdef OS_WIN32
+	if (rv != CKR_OK && sock->write_fd != -1) {
+		p11_message ("closing socket due to protocol failure");
+		close (sock->write_fd);
+		sock->write_fd = -1;
+	}
+#endif
 
 	sock->refs--;
 	assert (sock->refs > 0);
@@ -771,6 +830,191 @@ rpc_exec_connect (p11_rpc_client_vtable *vtable,
 	return CKR_OK;
 }
 
+#endif /* OS_UNIX */
+
+#ifdef OS_WIN32
+
+typedef struct {
+	p11_rpc_transport base;
+	p11_array *argv;
+	HANDLE pid;
+} rpc_exec;
+
+static void
+rpc_exec_wait_or_terminate (HANDLE pid)
+{
+	DWORD status;
+	int ret;
+	int i;
+
+
+	for (i = 0; i < 3 * 1000; i += 100) {
+		ret = WaitForSingleObject (pid, 10000);
+		if (ret == WAIT_OBJECT_0)
+			break;
+	}
+
+	if (ret != WAIT_OBJECT_0) {
+		p11_message ("process %p did not exit, terminating", pid);
+		if (!TerminateProcess (pid, SIGTERM))
+			p11_message ("couldn't terminate process %p", pid);
+		ret = WaitForSingleObject (pid, 0);
+	}
+
+	if (ret != WAIT_OBJECT_0) {
+		p11_message ("failed to wait for executed child: %p", pid);
+		status = 0;
+	} else if (!GetExitCodeProcess (pid, &status)) {
+		p11_message ("failed to get the exit status of %p", pid);
+	} else if (status == 0) {
+		p11_debug ("process %p exited with status 0", pid);
+	} else {
+		p11_message ("process %p exited with status %lu", pid, status);
+	}
+
+	CloseHandle (pid);
+}
+
+static void
+rpc_exec_disconnect (p11_rpc_client_vtable *vtable,
+                     void *fini_reserved)
+{
+	rpc_exec *rex = (rpc_exec *)vtable;
+
+	if (rex->base.socket)
+		rpc_socket_close (rex->base.socket);
+
+	if (rex->pid != INVALID_HANDLE_VALUE)
+		rpc_exec_wait_or_terminate (rex->pid);
+	rex->pid = INVALID_HANDLE_VALUE;
+
+	/* Do the common disconnect stuff */
+	rpc_transport_disconnect (vtable, fini_reserved);
+}
+
+static int
+set_cloexec_on_fd (int fd)
+{
+	HANDLE handle;
+
+	handle = (HANDLE) _get_osfhandle (fd);
+	if (!SetHandleInformation (handle, HANDLE_FLAG_INHERIT, 0))
+		return -1;
+
+	return 0;
+}
+
+static CK_RV
+rpc_exec_connect (p11_rpc_client_vtable *vtable,
+                  void *init_reserved)
+{
+	rpc_exec *rex = (rpc_exec *)vtable;
+	intptr_t pid = -1;
+	int pw[2] = { -1, -1 }, pr[2] = { -1, -1 };
+	int fds[2] = { -1, -1 };
+	CK_RV rv = CKR_OK;
+
+	p11_debug ("executing rpc transport: %s", (char *)rex->argv->elem[0]);
+
+	setvbuf (stdout, NULL, _IONBF, 0 );
+
+	if (_pipe (pw, 256, _O_BINARY) == -1 ||
+	    set_cloexec_on_fd (pw[1]) == -1) {
+		p11_message_err (errno, "failed to create pipe for remote");
+		rv = CKR_DEVICE_ERROR;
+		goto out;
+	}
+
+	if (_pipe (pr, 256, _O_BINARY) == -1 ||
+	    set_cloexec_on_fd (pr[0]) == -1) {
+		p11_message_err (errno, "failed to create pipe for remote");
+		rv = CKR_DEVICE_ERROR;
+		goto out;
+	}
+
+	/* Save the original stdin and stdout */
+	fds[0] = dup (STDIN_FILENO);
+	if (fds[0] == -1) {
+		p11_message_err (errno, "failed to duplicate stdin");
+		rv = CKR_DEVICE_ERROR;
+		goto out;
+	}
+	/* FIXME: Shouldn't we close STDIN_FILENO? */
+
+	fds[1] = dup (STDOUT_FILENO);
+	if (fds[1] == -1) {
+		p11_message_err (errno, "failed to duplicate stdout");
+		rv = CKR_DEVICE_ERROR;
+		goto out;
+	}
+	/* FIXME: Shouldn't we close STDOUT_FILENO */
+
+	/* Temporarily redirect pipe descriptors to stdin/stdout for child */
+	if (dup2 (pw[0], STDIN_FILENO) == -1 ||
+	    dup2 (pr[1], STDOUT_FILENO) == -1) {
+		p11_message_err (errno, "failed to duplicate child end of pipe");
+		rv = CKR_DEVICE_ERROR;
+		goto out;
+	}
+
+	pid = _spawnv (P_NOWAIT,
+		       rex->argv->elem[0],
+		       (const char * const *)rex->argv->elem);
+
+	if (pid == -1) {
+		p11_message_err (errno, "failed to spawn remote");
+		rv = CKR_DEVICE_ERROR;
+		goto out;
+	}
+
+	close (pw[0]);
+	pw[0] = -1;
+	close (pr[1]);
+	pr[1] = -1;
+
+	/* Restore the original stdin and stdout */
+	if (dup2 (fds[0], STDIN_FILENO) == -1 ||
+	    dup2 (fds[1], STDOUT_FILENO) == -1) {
+		p11_message_err (errno, "failed to restore file descriptors");
+		rv = CKR_DEVICE_ERROR;
+		goto out;
+	}
+
+	close (fds[0]);
+	fds[0] = -1;
+	close (fds[1]);
+	fds[1] = -1;
+
+	rex->pid = (HANDLE) pid;
+	rex->base.socket = rpc_socket_new (pr[0]);
+	return_val_if_fail (rex->base.socket != NULL, CKR_GENERAL_ERROR);
+	rex->base.socket->write_fd = pw[1];
+
+ out:
+	if (rv != CKR_OK) {
+		if (pid != -1) {
+			TerminateProcess ((HANDLE) pid, SIGTERM);
+			CloseHandle ((HANDLE) pid);
+		}
+		if (pw[0] != -1)
+			close (pw[0]);
+		if (pw[1] != -1)
+			close (pw[1]);
+		if (pr[0] != -1)
+			close (pr[0]);
+		if (pr[1] != -1)
+			close (pr[1]);
+		if (fds[0] != -1)
+			close (fds[0]);
+		if (fds[1] != -1)
+			close (fds[1]);
+	}
+
+	return rv;
+}
+
+#endif /* OS_WIN32 */
+
 static void
 rpc_exec_free (void *data)
 {
@@ -810,6 +1054,9 @@ rpc_exec_init (const char *remote,
 
 	p11_array_push (argv, NULL);
 	rex->argv = argv;
+#ifdef OS_WIN32
+	rex->pid = INVALID_HANDLE_VALUE;
+#endif
 
 	rex->base.vtable.connect = rpc_exec_connect;
 	rex->base.vtable.disconnect = rpc_exec_disconnect;
@@ -820,6 +1067,82 @@ rpc_exec_init (const char *remote,
 	return &rex->base;
 }
 
+#ifdef OS_UNIX
+
+typedef struct {
+	p11_rpc_transport base;
+	struct sockaddr_un sa;
+} rpc_unix;
+
+static CK_RV
+rpc_unix_connect (p11_rpc_client_vtable *vtable,
+		    void *init_reserved)
+{
+	rpc_unix *run = (rpc_unix *)vtable;
+	int fd;
+
+	fd = socket (AF_UNIX, SOCK_STREAM, 0);
+	if (fd < 0) {
+		p11_message_err (errno, "failed to create socket for remote");
+		return CKR_GENERAL_ERROR;
+	}
+
+	if (connect (fd, (struct sockaddr *)&run->sa, sizeof (run->sa)) < 0) {
+		p11_debug_err (errno, "failed to connect to socket");
+		close (fd);
+		return CKR_DEVICE_REMOVED;
+	}
+
+	run->base.socket = rpc_socket_new (fd);
+	return_val_if_fail (run->base.socket != NULL, CKR_GENERAL_ERROR);
+
+	return CKR_OK;
+}
+
+static void
+rpc_unix_disconnect (p11_rpc_client_vtable *vtable,
+                     void *fini_reserved)
+{
+	rpc_unix *run = (rpc_unix *)vtable;
+
+	if (run->base.socket)
+		rpc_socket_close (run->base.socket);
+
+	/* Do the common disconnect stuff */
+	rpc_transport_disconnect (vtable, fini_reserved);
+}
+
+static void
+rpc_unix_free (void *data)
+{
+	rpc_unix *run = data;
+	rpc_unix_disconnect (data, NULL);
+	rpc_transport_uninit (&run->base);
+	free (run);
+}
+
+static p11_rpc_transport *
+rpc_unix_init (const char *remote,
+	       const char *name)
+{
+	rpc_unix *run;
+
+	run = calloc (1, sizeof (rpc_unix));
+	return_val_if_fail (run != NULL, NULL);
+
+	memset (&run->sa, 0, sizeof (run->sa));
+	run->sa.sun_family = AF_UNIX;
+	snprintf (run->sa.sun_path, sizeof (run->sa.sun_path), "%s", remote);
+
+	run->base.vtable.connect = rpc_unix_connect;
+	run->base.vtable.disconnect = rpc_unix_disconnect;
+	run->base.vtable.transport = rpc_transport_buffer;
+	rpc_transport_init (&run->base, name, rpc_unix_free);
+
+	p11_debug ("initialized rpc socket: %s", remote);
+	return &run->base;
+}
+
 #endif /* OS_UNIX */
 
 p11_rpc_transport *
@@ -833,15 +1156,20 @@ p11_rpc_transport_new (p11_virtual *virt,
 	return_val_if_fail (remote != NULL, NULL);
 	return_val_if_fail (name != NULL, NULL);
 
-#ifdef OS_WIN32
-	p11_message ("Windows not yet supported for remote");
-	return NULL;
-#endif
-
 	/* This is a command we can execute */
 	if (remote[0] == '|') {
 		rpc = rpc_exec_init (remote + 1, name);
 
+#ifdef OS_UNIX
+	} else if (strncmp (remote, "unix:path=/", 11) == 0) {
+		/* Only absolute path is supported */
+		char *path;
+
+		path = p11_path_decode (remote + 10);
+		return_val_if_fail (path != NULL, NULL);
+		rpc = rpc_unix_init (path, name);
+		free (path);
+#endif /* OS_UNIX */
 	} else {
 		p11_message ("remote not supported: %s", remote);
 		return NULL;
